;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Abstract Syntax definitions.

; option =
(define-const-structure (Some _))
(define-const-structure (None))
(define None (make-None))
(define Some make-Some)

; A plural type name means a list of the type.

; exp =
(define-const-structure (And    exps))
(define-const-structure (App    exp exps))
(define-const-structure (Begin  exps))
(define-const-structure (Const  val pred))         ; pred is a Name
(define-const-structure (If     exp1 exp2 exp3))
(define-const-structure (Lam    names body))
(define-const-structure (Let    binds body))
(define-const-structure (Let*   binds body))
(define-const-structure (Letr   binds body))
(define-const-structure (Or     exps))
(define-const-structure (Prim   name))             ; gone after binding
(define-const-structure (Delay  exp))
(define-const-structure (Set!   (! name) exp))
(define-const-structure (Var    (! name)))
(define-const-structure (Vlam   names name body))
(define-const-structure (Match  exp mclauses))
(define-const-structure (Record binds))
(define-const-structure (Field  name exp))
(define-const-structure (Cast   type exp))

; body =
(define-const-structure (Body   defs exps))

; bind =
(define-const-structure (Bind   name exp))

; mclause =
(define-const-structure (Mclause pat body fail))   ; fail is #f or a Name

; pat =
(define-const-structure (Pvar   name))     ; pattern variable
(define-const-structure (Pany))            ; _
(define-const-structure (Pelse))           ; else
(define-const-structure (Pconst name pred)); eg. 1, "hi", but not #t, #f, or ()
(define-const-structure (Pobj   name pats)); composite, eg. (cons p1 p2)
(define-const-structure (Ppred  name))     ; eg. (? exp)
(define-const-structure (Pand   pats))     ; (length pats) >= 1
(define-const-structure (Pnot   pat))      ; eg. (not pat)

; definitions =
(define-const-structure (Define name (! exp)))
(define-const-structure (Defstruct tag args make pred get set getn setn mutable))

; unsupported definitions
(define-const-structure (Datatype _))
; _ = (list (cons (list symbol) (list (Variant name (list type)))))))
(define-const-structure (Variant con pred arg-types))

; name =
(define-structure
  (Name name         ; name
        ty           ; type
        timestamp    ; number, or context if unbound
        occ          ; # of occurrences
        mutated      ; set!'ed?
        gdef         ; user defined global name
        primitive    ; primitive
        struct       ; built by define-structure
        pure         ; pure fn, bool or (list bool) for constructors
        predicate    ; type template if predicate
        variant      ; type template for variant
        selector     ; match template if selector
        ))

; Decorations inserted by the type checker.
(define-structure (Type   ty exp))
(define-const-structure (Shape  _ _))
(define-const-structure (Check  _ _))

