;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Routines for displaying checks

(define check
  (lambda (file)
    (output-checked file '() type-check?)))

(define profcheck
  (lambda (file)
    (output-checked #f '() type-check?)
    (output-checked #f (make-counters total-possible) type-check?)))

(define fullcheck
  (lambda (file)
    (let ((check? (lambda (_) #t)))
      (output-checked #f '() check?)
      (output-checked #f (make-counters total-possible) check?))))

(define make-counters
  (lambda (n)
    (let* ((init `(define CHECK-COUNTERS (make-vector ,n 0)))
           (sum '(define CHECK-TOTAL
                   (lambda ()
                     (let ((foldr
                            (lambda (f i l)
                              (recur loop ((l l))
                                (match l
                                  (() i)
                                  ((x . y) (f x (loop y))))))))
                       (foldr + 0 (vector->list CHECK-COUNTERS))))))
           (incr '(extend-syntax (CHECK-INCREMENT-COUNTER)
                    ((CHECK-INCREMENT-COUNTER c)
                     (vector-set! CHECK-COUNTERS c
                       (+ 1 (vector-ref CHECK-COUNTERS c)))))))
      (list init sum incr))))

(define output-checked
  (lambda (file header check-test)
    (set! summary '())
    (set! total-possible 0)
    (set! total-cast 0)
    (set! total-err 0)
    (set! total-any 0)
    (let ((doit (lambda ()
                  (when (string? file)
                    (printf ";; Generated by Soft Scheme ~a~%" st:version)
                    (printf ";; (st:control")
                    (for-each (lambda (x) (printf " '~a" x)) (show-controls))
                    (printf ")~%")
                    (unless (= 0 n-unbound)
                      (printf ";; CAUTION: ~a unbound references, this code is not safe~%" n-unbound))
                    (printf "~%")
                    (for-each pretty-print header))
                  (for-each
                    (lambda (exp)
                      (match exp
                        (($ Define x _)
                         (set! n-possible 0)
                         (set! n-clash 0)
                         (set! n-err 0)
                         (set! n-match 0)
                         (set! n-inexhaust 0)
                         (set! n-prim 0)
                         (set! n-lam 0)
                         (set! n-app 0)
                         (set! n-field 0)
                         (set! n-cast 0)
                         (if file
                             (pretty-print (pcheck exp check-test))
                             (pcheck exp check-test))
                         (make-summary-line x)
                         (set! total-possible (+ total-possible n-possible))
                         (set! total-cast (+ total-cast n-cast))
                         (set! total-err (+ total-err n-err))
                         (set! total-any (+ total-any
                                            n-match
                                            n-inexhaust
                                            n-prim
                                            n-lam
                                            n-app
                                            n-field
                                            n-cast)))
                        (_
                         (when file
                           (pretty-print (pcheck exp check-test))))))
                    tree)
                  (when (string? file)
                    (newline)
                    (newline)
                    (print-summary "; ")))))
      (if (string? file)
          (begin
            (delete-file file)
            (with-output-to-file file doit))
          (doit)))))

(define total-possible 0)
(define total-err 0)
(define total-cast 0)
(define total-any 0)

(define n-possible 0)
(define n-clash 0)
(define n-err 0)
(define n-match 0)
(define n-inexhaust 0)
(define n-prim 0)
(define n-lam 0)
(define n-app 0)
(define n-field 0)
(define n-cast 0)

(define summary '())

(define make-summary-line
  (lambda (x)
    (let ((total (+ n-match n-inexhaust n-prim n-lam n-app n-field n-cast)))
      (unless (= 0 total)
        (let* ((s (sprintf "~a~a " (padr (pname x) 16) (padl total 2)))
               (s (cond ((< 0 n-inexhaust)
                         (sprintf "~a (~a match ~a inexhaust)"
                           s n-match n-inexhaust))
                        ((< 0 n-match)
                         (sprintf "~a (~a match)" s n-match))
                        (else
                         s)))
               (s (if (< 0 n-prim)
                      (sprintf "~a (~a prim)" s n-prim)
                      s))
               (s (if (< 0 n-field)
                      (sprintf "~a (~a field)" s n-field)
                      s))
               (s (if (< 0 n-lam)
                      (sprintf "~a (~a lambda)" s n-lam)
                      s))
               (s (if (< 0 n-app)
                      (sprintf "~a (~a ap)" s n-app)
                      s))
               (s (if (< 0 n-err)
                      (sprintf "~a (~a ERROR)" s n-err)
                      s))
               (s (if (< 0 n-cast)
                      (sprintf "~a (~a TYPE)" s n-cast)
                      s)))
          (set! summary (cons s summary)))))))

(define print-summary
  (lambda (hdr)
    (for-each
      (lambda (s) (printf "~a~a~%" hdr s))
      (reverse summary))
    (printf "~a~a~a " hdr (padr "TOTAL CHECKS" 16) (padl total-any 2))
    (printf " (of ~s is ~s%)"
      total-possible
      (if (= 0 total-possible)
          0
          (string->number
            (chop-number
              (exact->inexact (* (/ total-any total-possible) 100))
              4))))
    (when (< 0 total-err)
      (printf " (~s ERROR)" total-err))
    (when (< 0 total-cast)
      (printf " (~s TYPE)" total-cast))
    (printf "~%")))

;; pad on left
(define padl
  (lambda (arg n)
    (let ((s (sprintf "~a" arg)))
      (recur loop ((s s))
        (if (< (string-length s) n)
            (loop (string-append " " s))
            s)))))

;; pad on right
(define padr
  (lambda (arg n)
    (let ((s (sprintf "~a" arg)))
      (recur loop ((s s))
        (if (< (string-length s) n)
            (loop (string-append s " "))
            s)))))

(define chop-number
  (lambda (x n)
    (substring (sprintf "~s00000000000000000000" x) 0 (- n 1))))

(define pcheck
  (lambda (ex check-test)
    (unparse ex
      (lambda (e pexpr)
        (match e
          ((and z ($ Type _ ($ Check inf ($ Var x))))
           (cond ((Name-primitive x)
                  (set! n-possible (+ 1 n-possible))
                  (match (check-test inf)
                    (#f
                      (pname x))
                    ('DEF
                      (set! n-err (+ 1 n-err))
                      (set! n-prim (+ 1 n-prim))
                      `(,(symbol-append "CHECK-" (pname x)) ,(tree-index z)
                        ',(string->symbol "ERROR")))
                    (_
                     (set! n-prim (+ 1 n-prim))
                     `(,(symbol-append "CHECK-" (pname x)) ,(tree-index z)))))
                 ((Name-unbound? x)
                  `(CHECK-BOUND ,(pname x)))
                 (else
                  (if (check-test inf)
                      (begin
                        (set! n-clash (+ 1 n-clash))
                        `(,(string->symbol "CLASH") ,(pname x) ,(tree-index z)))
                      (pname x)))))
          ((and z ($ Type _ ($ Check inf (and m ($ Lam x e1)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             ('DEF
               (set! n-err (+ 1 n-err))
               (set! n-lam (+ 1 n-lam))
               `(,(string->symbol "CHECK-lambda")
                 (,(tree-index z) ',(string->symbol "ERROR"))
                 ,(map pname x) ,@(pexpr e1)))
             (_
               (set! n-lam (+ 1 n-lam))
              `(,(string->symbol "CHECK-lambda") (,(tree-index z))
                ,(map pname x) ,@(pexpr e1)))))
          ((and z ($ Type _ ($ Check inf (and m ($ Vlam x rest e1)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             ('DEF
               (set! n-err (+ 1 n-err))
               (set! n-lam (+ 1 n-lam))
               `(,(string->symbol "CHECK-lambda")
                 (,(tree-index z) ',(string->symbol "ERROR"))
                 ,(append (map pname x) (pname rest)) ,@(pexpr e1)))
             (_
              (set! n-lam (+ 1 n-lam))
              `(,(string->symbol "CHECK-lambda") (,(tree-index z))
                ,(append (map pname x) (pname rest)) ,@(pexpr e1)))))
          ((and z ($ Type _ ($ Check inf (and m ($ App e1 args)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             ('DEF
               (set! n-err (+ 1 n-err))
               (set! n-app (+ 1 n-app))
               `(,(string->symbol "CHECK-ap")
                 (,(tree-index z) ',(string->symbol "ERROR"))
                 ,(pexpr e1) ,@(map pexpr args)))
             (_ 
              (set! n-app (+ 1 n-app))
              (let ((p1 (pexpr e1)))
                `(,(string->symbol "CHECK-ap") (,(tree-index z))
                  ,p1 ,@(map pexpr args))))))
          ((and z ($ Type _ ($ Check inf (and m ($ Field x e1)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             ('DEF
               (set! n-err (+ 1 n-err))
               (set! n-field (+ 1 n-field))
               `(,(string->symbol "CHECK-field")
                 (,(tree-index z) ',(string->symbol "ERROR"))
                 ,x ,(pexpr e1)))
             (_
               (set! n-field (+ 1 n-field))
              `(,(string->symbol "CHECK-field") (,(tree-index z))
                ,x ,(pexpr e1)))))
          ((and z ($ Type _ ($ Check inf (and m ($ Cast (x . _) e1)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             (_
              (set! n-cast (+ 1 n-cast))
              `(,(string->symbol "CHECK-:") (,(tree-index z))
                ,x ,(pexpr e1)))))
          ((and z ($ Type _ ($ Check inf (and m ($ Match e1 clauses)))))
           (set! n-possible (+ 1 n-possible))
           (match (check-test inf)
             (#f (pexpr m))
             (inx (let* ((pclause
                          (match-lambda
                            (($ Mclause p exp fail)
                             (if fail
                                 `(,(ppat p) (=> ,(pname fail)) ,@(pexpr exp))
                                 `(,(ppat p) ,@(pexpr exp))))))
                         (p1 (pexpr e1)))
                    (if (eq? 'INEXHAUST inx)
                        (begin
                          (set! n-inexhaust (+ 1 n-inexhaust))
                          `(,(string->symbol "CHECK-match")
                            (,(tree-index z) ,(string->symbol "INEXHAUST"))
                            ,p1 ,@(mapLR pclause clauses)))
                        (begin
                          (set! n-match (+ 1 n-match))
                          `(,(string->symbol "CHECK-match")
                            (,(tree-index z))
                            ,p1 ,@(mapLR pclause clauses)))))))))))))

(define tree-index-list '())

(define reinit-output!
  (lambda ()
    (set! tree-index-list '())))

(define tree-index
  (lambda (syntax)
    (match (assq syntax tree-index-list)
      (#f (let ((n (length tree-index-list)))
            (set! tree-index-list (cons (cons syntax n) tree-index-list))
            n))
      ((_ . n) n))))

(define tree-unindex
  (lambda (n)
    (let ((max (length tree-index-list)))
      (when (<= max n)
        (use-error "Invalid CHECK number ~a" n))
      (car (list-ref tree-index-list (- (- max 1) n))))))
