;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Soft Scheme - Copyright (c) 1993,1994 by Andrew K. Wright
;; See LICENSE for a full description of the license terms.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Things for running soft typed code at speed.
;; Load this after checklib.scm.

(defmacro CHECK-BOUND args
  (match args
    ((var) var)
    (x (st:failure `(CHECK-BOUND ,@x) "syntax-error"))))

(defmacro CLASH args
  (match args
    ((name info ...) name)
    (x (#%error `(CLASH ,@x) "syntax error"))))

(defmacro CHECK-lambda args
  (match args
    (((id info ...) args body0 body ...) `(lambda ,args ,body0 ,@body))
    (x (#%error `(CHECK-lambda ,@x) "syntax error"))))

(defmacro CHECK-ap args
  (match args
    (((id info ...) (? symbol? f) args ...)
     `(if (#%procedure? ,f)
          (,f ,@args)
          (#%error '(CHECK-ap ,id ,@info) "not a procedure: ~a" ,f)))
    (((id info ...) f args ...)
     (let ((g (gensym)))
       `(let ((,g ,f))
          (if (#%procedure? ,g)
              (,g ,@args)
              (#%error '(CHECK-ap ,id ,@info) "not a procedure: ~a" ,g)))))
    (x (#%error `(CHECK-ap ,@x) "syntax error"))))

(defmacro CHECK-field args
  (match args
    (((id info ...) (? symbol? f) exp)
     `(field ,f ,exp))
    (x (#%error `(CHECK-field ,@x) "syntax error"))))

(defmacro CHECK-match args
  (match args
    (((id info ...) exp clause ...)
     `(match ,exp ,@clause))))

(defmacro CHECK-: args
  (match args
    (((id info ...) typeexp exp)
     `(#%error '(CHECK-: ,id ,@info) "static type annotation reached"))
    (x (#%error `(CHECK-: ,@x) "syntax error"))))

    ; predicates
(extend-syntax (CHECK-number?) ((CHECK-number? id ...) #2%number?))
(extend-syntax (CHECK-null?) ((CHECK-null? id ...) #2%null?))
(extend-syntax (CHECK-char?) ((CHECK-char? id ...) #2%char?))
(extend-syntax (CHECK-symbol?) ((CHECK-symbol? id ...) #2%symbol?))
(extend-syntax (CHECK-string?) ((CHECK-string? id ...) #2%string?))
(extend-syntax (CHECK-vector?) ((CHECK-vector? id ...) #2%vector?))
(extend-syntax (CHECK-box?) ((CHECK-box? id ...) #2%box?))
(extend-syntax (CHECK-pair?) ((CHECK-pair? id ...) #2%pair?))
(extend-syntax (CHECK-procedure?) ((CHECK-procedure? id ...) #2%procedure?))
(extend-syntax (CHECK-eof-object?) ((CHECK-eof-object? id ...) #2%eof-object?))
(extend-syntax (CHECK-input-port?) ((CHECK-input-port? id ...) #2%input-port?))
(extend-syntax (CHECK-output-port?) ((CHECK-output-port? id ...) #2%output-port?))
(extend-syntax (CHECK-true-object?) ((CHECK-true-object? id ...) true-object?))
(extend-syntax (CHECK-false-object?) ((CHECK-false-object? id ...) false-object?))
(extend-syntax (CHECK-boolean?) ((CHECK-boolean? id ...) #2%boolean?))
(extend-syntax (CHECK-list?) ((CHECK-list? id ...) #2%list?))

    ; booleans
(extend-syntax (CHECK-not) ((CHECK-not id ...) #2%not))

    ; equivalence predicates
(extend-syntax (CHECK-eqv?) ((CHECK-eqv? id ...) #2%eqv?))
(extend-syntax (CHECK-eq?) ((CHECK-eq? id ...) #2%eq?))
(extend-syntax (CHECK-equal?) ((CHECK-equal? id ...) #2%equal?))

    ; pairs and lists
(extend-syntax (CHECK-cons) ((CHECK-cons id ...) #2%cons))
(extend-syntax (CHECK-car) ((CHECK-car id ...) #2%car))
(extend-syntax (CHECK-cdr) ((CHECK-cdr id ...) #2%cdr))
(extend-syntax (CHECK-caar) ((CHECK-caar id ...) #2%caar))
(extend-syntax (CHECK-cadr) ((CHECK-cadr id ...) #2%cadr))
(extend-syntax (CHECK-cdar) ((CHECK-cdar id ...) #2%cdar))
(extend-syntax (CHECK-cddr) ((CHECK-cddr id ...) #2%cddr))
(extend-syntax (CHECK-caaar) ((CHECK-caaar id ...) #2%caaar))
(extend-syntax (CHECK-caadr) ((CHECK-caadr id ...) #2%caadr))
(extend-syntax (CHECK-cadar) ((CHECK-cadar id ...) #2%cadar))
(extend-syntax (CHECK-caddr) ((CHECK-caddr id ...) #2%caddr))
(extend-syntax (CHECK-cdaar) ((CHECK-cdaar id ...) #2%cdaar))
(extend-syntax (CHECK-cdadr) ((CHECK-cdadr id ...) #2%cdadr))
(extend-syntax (CHECK-cddar) ((CHECK-cddar id ...) #2%cddar))
(extend-syntax (CHECK-cdddr) ((CHECK-cdddr id ...) #2%cdddr))
(extend-syntax (CHECK-caaaar) ((CHECK-caaaar id ...) #2%caaaar))
(extend-syntax (CHECK-caaadr) ((CHECK-caaadr id ...) #2%caaadr))
(extend-syntax (CHECK-caadar) ((CHECK-caadar id ...) #2%caadar))
(extend-syntax (CHECK-caaddr) ((CHECK-caaddr id ...) #2%caaddr))
(extend-syntax (CHECK-cadaar) ((CHECK-cadaar id ...) #2%cadaar))
(extend-syntax (CHECK-cadadr) ((CHECK-cadadr id ...) #2%cadadr))
(extend-syntax (CHECK-caddar) ((CHECK-caddar id ...) #2%caddar))
(extend-syntax (CHECK-cadddr) ((CHECK-cadddr id ...) #2%cadddr))
(extend-syntax (CHECK-cdaaar) ((CHECK-cdaaar id ...) #2%cdaaar))
(extend-syntax (CHECK-cdaadr) ((CHECK-cdaadr id ...) #2%cdaadr))
(extend-syntax (CHECK-cdadar) ((CHECK-cdadar id ...) #2%cdadar))
(extend-syntax (CHECK-cdaddr) ((CHECK-cdaddr id ...) #2%cdaddr))
(extend-syntax (CHECK-cddaar) ((CHECK-cddaar id ...) #2%cddaar))
(extend-syntax (CHECK-cddadr) ((CHECK-cddadr id ...) #2%cddadr))
(extend-syntax (CHECK-cdddar) ((CHECK-cdddar id ...) #2%cdddar))
(extend-syntax (CHECK-cddddr) ((CHECK-cddddr id ...) #2%cddddr))
(extend-syntax (CHECK-set-car!) ((CHECK-set-car! id ...) #2%set-car!))
(extend-syntax (CHECK-set-cdr!) ((CHECK-set-cdr! id ...) #2%set-cdr!))
(extend-syntax (CHECK-list) ((CHECK-list id ...) #2%list))
(extend-syntax (CHECK-length) ((CHECK-length id ...) #2%length))
(extend-syntax (CHECK-append) ((CHECK-append id ...) #2%append))
(extend-syntax (CHECK-reverse) ((CHECK-reverse id ...) #2%reverse))
(extend-syntax (CHECK-list-tail) ((CHECK-list-tail id ...) #2%list-tail))
(extend-syntax (CHECK-list-ref) ((CHECK-list-ref id ...) #2%list-ref))
(extend-syntax (CHECK-memq) ((CHECK-memq id ...) #2%memq))
(extend-syntax (CHECK-memv) ((CHECK-memv id ...) #2%memv))
(extend-syntax (CHECK-member) ((CHECK-member id ...) #2%member))
(extend-syntax (CHECK-assq) ((CHECK-assq id ...) #2%assq))
(extend-syntax (CHECK-assv) ((CHECK-assv id ...) #2%assv))
(extend-syntax (CHECK-assoc) ((CHECK-assoc id ...) #2%assoc))

    ; symbols
(extend-syntax (CHECK-symbol->string) ((CHECK-symbol->string id ...) #2%symbol->string))
(extend-syntax (CHECK-string->symbol) ((CHECK-string->symbol id ...) #2%string->symbol))

    ; numbers
(extend-syntax (CHECK-complex?) ((CHECK-complex? id ...) #2%complex?))
(extend-syntax (CHECK-real?) ((CHECK-real? id ...) #2%real?))
(extend-syntax (CHECK-rational?) ((CHECK-rational? id ...) #2%rational?))
(extend-syntax (CHECK-integer?) ((CHECK-integer? id ...) #2%integer?))
(extend-syntax (CHECK-exact?) ((CHECK-exact? id ...) #2%exact?))
(extend-syntax (CHECK-inexact?) ((CHECK-inexact? id ...) #2%inexact?))
(extend-syntax (CHECK-=) ((CHECK-= id ...) #2%=))
(extend-syntax (CHECK-<) ((CHECK-< id ...) #2%<))
(extend-syntax (CHECK->) ((CHECK-> id ...) #2%>))
(extend-syntax (CHECK-<=) ((CHECK-<= id ...) #2%<=))
(extend-syntax (CHECK->=) ((CHECK->= id ...) #2%>=))
(extend-syntax (CHECK-zero?) ((CHECK-zero? id ...) #2%zero?))
(extend-syntax (CHECK-positive?) ((CHECK-positive? id ...) #2%positive?))
(extend-syntax (CHECK-negative?) ((CHECK-negative? id ...) #2%negative?))
(extend-syntax (CHECK-odd?) ((CHECK-odd? id ...) #2%odd?))
(extend-syntax (CHECK-even?) ((CHECK-even? id ...) #2%even?))
(extend-syntax (CHECK-max) ((CHECK-max id ...) #2%max))
(extend-syntax (CHECK-min) ((CHECK-min id ...) #2%min))
(extend-syntax (CHECK-+) ((CHECK-+ id ...) #2%+))
(extend-syntax (CHECK-*) ((CHECK-* id ...) #2%*))
(extend-syntax (CHECK--) ((CHECK-- id ...) #2%-))
(extend-syntax (CHECK-/) ((CHECK-/ id ...) #2%/))
(extend-syntax (CHECK-abs) ((CHECK-abs id ...) #2%abs))
(extend-syntax (CHECK-quotient) ((CHECK-quotient id ...) #2%quotient))
(extend-syntax (CHECK-remainder) ((CHECK-remainder id ...) #2%remainder))
(extend-syntax (CHECK-modulo) ((CHECK-modulo id ...) #2%modulo))
(extend-syntax (CHECK-gcd) ((CHECK-gcd id ...) #2%gcd))
(extend-syntax (CHECK-lcm) ((CHECK-lcm id ...) #2%lcm))
(extend-syntax (CHECK-numerator) ((CHECK-numerator id ...) #2%numerator))
(extend-syntax (CHECK-denominator) ((CHECK-denominator id ...) #2%denominator))
(extend-syntax (CHECK-floor) ((CHECK-floor id ...) #2%floor))
(extend-syntax (CHECK-ceiling) ((CHECK-ceiling id ...) #2%ceiling))
(extend-syntax (CHECK-truncate) ((CHECK-truncate id ...) #2%truncate))
(extend-syntax (CHECK-round) ((CHECK-round id ...) #2%round))
(extend-syntax (CHECK-rationalize) ((CHECK-rationalize id ...) #2%rationalize))
(extend-syntax (CHECK-exp) ((CHECK-exp id ...) #2%exp))
(extend-syntax (CHECK-log) ((CHECK-log id ...) #2%log))
(extend-syntax (CHECK-sin) ((CHECK-sin id ...) #2%sin))
(extend-syntax (CHECK-cos) ((CHECK-cos id ...) #2%cos))
(extend-syntax (CHECK-tan) ((CHECK-tan id ...) #2%tan))
(extend-syntax (CHECK-asin) ((CHECK-asin id ...) #2%asin))
(extend-syntax (CHECK-acos) ((CHECK-acos id ...) #2%acos))
(extend-syntax (CHECK-atan) ((CHECK-atan id ...) #2%atan))
(extend-syntax (CHECK-sqrt) ((CHECK-sqrt id ...) #2%sqrt))
(extend-syntax (CHECK-expt) ((CHECK-expt id ...) #2%expt))
(extend-syntax (CHECK-make-rectangular) ((CHECK-make-rectangular id ...) #2%make-rectangular))
(extend-syntax (CHECK-make-polar) ((CHECK-make-polar id ...) #2%make-polar))
(extend-syntax (CHECK-real-part) ((CHECK-real-part id ...) #2%real-part))
(extend-syntax (CHECK-imag-part) ((CHECK-imag-part id ...) #2%imag-part))
(extend-syntax (CHECK-magnitude) ((CHECK-magnitude id ...) #2%magnitude))
(extend-syntax (CHECK-angle) ((CHECK-angle id ...) #2%angle))
(extend-syntax (CHECK-exact->inexact) ((CHECK-exact->inexact id ...) #2%exact->inexact))
(extend-syntax (CHECK-inexact->exact) ((CHECK-inexact->exact id ...) #2%inexact->exact))
(extend-syntax (CHECK-number->string) ((CHECK-number->string id ...) #2%number->string))
(extend-syntax (CHECK-string->number) ((CHECK-string->number id ...) #2%string->number))

    ; characters
(extend-syntax (CHECK-char=?) ((CHECK-char=? id ...) #2%char=?))
(extend-syntax (CHECK-char<?) ((CHECK-char<? id ...) #2%char<?))
(extend-syntax (CHECK-char>?) ((CHECK-char>? id ...) #2%char>?))
(extend-syntax (CHECK-char<=?) ((CHECK-char<=? id ...) #2%char<=?))
(extend-syntax (CHECK-char>=?) ((CHECK-char>=? id ...) #2%char>=?))
(extend-syntax (CHECK-char-ci=?) ((CHECK-char-ci=? id ...) #2%char-ci=?))
(extend-syntax (CHECK-char-ci<?) ((CHECK-char-ci<? id ...) #2%char-ci<?))
(extend-syntax (CHECK-char-ci>?) ((CHECK-char-ci>? id ...) #2%char-ci>?))
(extend-syntax (CHECK-char-ci<=?) ((CHECK-char-ci<=? id ...) #2%char-ci<=?))
(extend-syntax (CHECK-char-ci>=?) ((CHECK-char-ci>=? id ...) #2%char-ci>=?))
(extend-syntax (CHECK-char-alphabetic?) ((CHECK-char-alphabetic? id ...) #2%char-alphabetic?))
(extend-syntax (CHECK-char-numeric?) ((CHECK-char-numeric? id ...) #2%char-numeric?))
(extend-syntax (CHECK-char-whitespace?) ((CHECK-char-whitespace? id ...) #2%char-whitespace?))
(extend-syntax (CHECK-char-upper-case?) ((CHECK-char-upper-case? id ...) #2%char-upper-case?))
(extend-syntax (CHECK-char-lower-case?) ((CHECK-char-lower-case? id ...) #2%char-lower-case?))
(extend-syntax (CHECK-char->integer) ((CHECK-char->integer id ...) #2%char->integer))
(extend-syntax (CHECK-integer->char) ((CHECK-integer->char id ...) #2%integer->char))
(extend-syntax (CHECK-char-upcase) ((CHECK-char-upcase id ...) #2%char-upcase))
(extend-syntax (CHECK-char-downcase) ((CHECK-char-downcase id ...) #2%char-downcase))

    ; strings
(extend-syntax (CHECK-make-string) ((CHECK-make-string id ...) #2%make-string))
(extend-syntax (CHECK-string) ((CHECK-string id ...) #2%string))
(extend-syntax (CHECK-string-length) ((CHECK-string-length id ...) #2%string-length))
(extend-syntax (CHECK-string-ref) ((CHECK-string-ref id ...) #2%string-ref))
(extend-syntax (CHECK-string-set!) ((CHECK-string-set! id ...) #2%string-set!))
(extend-syntax (CHECK-string=?) ((CHECK-string=? id ...) #2%string=?))
(extend-syntax (CHECK-string<?) ((CHECK-string<? id ...) #2%string<?))
(extend-syntax (CHECK-string>?) ((CHECK-string>? id ...) #2%string>?))
(extend-syntax (CHECK-string<=?) ((CHECK-string<=? id ...) #2%string<=?))
(extend-syntax (CHECK-string>=?) ((CHECK-string>=? id ...) #2%string>=?))
(extend-syntax (CHECK-string-ci=?) ((CHECK-string-ci=? id ...) #2%string-ci=?))
(extend-syntax (CHECK-string-ci<?) ((CHECK-string-ci<? id ...) #2%string-ci<?))
(extend-syntax (CHECK-string-ci>?) ((CHECK-string-ci>? id ...) #2%string-ci>?))
(extend-syntax (CHECK-string-ci<=?) ((CHECK-string-ci<=? id ...) #2%string-ci<=?))
(extend-syntax (CHECK-string-ci>=?) ((CHECK-string-ci>=? id ...) #2%string-ci>=?))
(extend-syntax (CHECK-substring) ((CHECK-substring id ...) #2%substring))
(extend-syntax (CHECK-string-append) ((CHECK-string-append id ...) #2%string-append))
(extend-syntax (CHECK-string->list) ((CHECK-string->list id ...) #2%string->list))
(extend-syntax (CHECK-list->string) ((CHECK-list->string id ...) #2%list->string))
(extend-syntax (CHECK-string-copy) ((CHECK-string-copy id ...) #2%string-copy))
(extend-syntax (CHECK-string-fill!) ((CHECK-string-fill! id ...) #2%string-fill!))

    ; vectors
(extend-syntax (CHECK-make-vector)
  ((CHECK-make-vector id ...)
   (lambda (n init) (#2%make-vector n init))))
(extend-syntax (CHECK-vector) ((CHECK-vector id ...) #2%vector))
(extend-syntax (CHECK-vector-length) ((CHECK-vector-length id ...) #2%vector-length))
(extend-syntax (CHECK-vector-ref) ((CHECK-vector-ref id ...) #2%vector-ref))
(extend-syntax (CHECK-vector-set!) ((CHECK-vector-set! id ...) #2%vector-set!))
(extend-syntax (CHECK-vector->list) ((CHECK-vector->list id ...) #2%vector->list))
(extend-syntax (CHECK-list->vector) ((CHECK-list->vector id ...) #2%list->vector))
(extend-syntax (CHECK-vector-fill!) ((CHECK-vector-fill! id ...) #2%vector-fill!))

    ; control features
(extend-syntax (CHECK-apply) ((CHECK-apply id ...) #2%apply))
(extend-syntax (CHECK-map) ((CHECK-map id ...) #2%map))
(extend-syntax (CHECK-for-each) ((CHECK-for-each id ...) #2%for-each))
(extend-syntax (CHECK-force) ((CHECK-force id ...) #2%force))
(extend-syntax (CHECK-call-with-current-continuation) ((CHECK-call-with-current-continuation id ...) #2%call-with-current-continuation))

    ; input and output
(extend-syntax (CHECK-call-with-input-file) ((CHECK-call-with-input-file id ...) #2%call-with-input-file))
(extend-syntax (CHECK-call-with-output-file) ((CHECK-call-with-output-file id ...) #2%call-with-output-file))
(extend-syntax (CHECK-input-port?) ((CHECK-input-port? id ...) #2%input-port?))
(extend-syntax (CHECK-output-port?) ((CHECK-output-port? id ...) #2%output-port?))
(extend-syntax (CHECK-current-input-port) ((CHECK-current-input-port id ...) #2%current-input-port))
(extend-syntax (CHECK-current-output-port) ((CHECK-current-output-port id ...) #2%current-output-port))
(extend-syntax (CHECK-with-input-from-file) ((CHECK-with-input-from-file id ...) #2%with-input-from-file))
(extend-syntax (CHECK-with-output-to-file) ((CHECK-with-output-to-file id ...) #2%with-output-to-file))
(extend-syntax (CHECK-open-input-file) ((CHECK-open-input-file id ...) #2%open-input-file))
(extend-syntax (CHECK-open-output-file) ((CHECK-open-output-file id ...) #2%open-output-file))
(extend-syntax (CHECK-close-input-port) ((CHECK-close-input-port id ...) #2%close-input-port))
(extend-syntax (CHECK-close-output-port) ((CHECK-close-output-port id ...) #2%close-output-port))
(extend-syntax (CHECK-read) ((CHECK-read id ...) #2%read))
(extend-syntax (CHECK-read-char) ((CHECK-read-char id ...) #2%read-char))
(extend-syntax (CHECK-peek-char) ((CHECK-peek-char id ...) #2%peek-char))
(extend-syntax (CHECK-char-ready?) ((CHECK-char-ready? id ...) #2%char-ready?))
(extend-syntax (CHECK-write) ((CHECK-write id ...) #2%write))
(extend-syntax (CHECK-display) ((CHECK-display id ...) #2%display))
(extend-syntax (CHECK-newline) ((CHECK-newline id ...) #2%newline))
(extend-syntax (CHECK-write-char) ((CHECK-write-char id ...) #2%write-char))

    ; system interface
(extend-syntax (CHECK-load) ((CHECK-load id ...) #2%load))
(extend-syntax (CHECK-transcript-on) ((CHECK-transcript-on id ...) #2%transcript-on))
(extend-syntax (CHECK-transcript-off) ((CHECK-transcript-off id ...) #2%transcript-off))

    ; non-R4RS extensions

    ; misc
(extend-syntax (CHECK-symbol-append) ((CHECK-symbol-append id ...) symbol-append))
(extend-syntax (CHECK-box) ((CHECK-box id ...) #2%box))
(extend-syntax (CHECK-unbox) ((CHECK-unbox id ...) #2%unbox))
(extend-syntax (CHECK-set-box!) ((CHECK-set-box! id ...) #2%set-box!))
(extend-syntax (CHECK-void) ((CHECK-void id ...) #2%void))
(extend-syntax (CHECK-make-module) ((CHECK-make-module id ...) make-module))
; raise never needs a CHECK

    ; for match
(extend-syntax (CHECK-match:error) ((CHECK-match:error id ...) match:error))

    ; for forms that return no useful value
(extend-syntax (CHECK-should-never-reach) ((CHECK-should-never-reach id ...) should-never-reach))

    ; constant vectors
(extend-syntax (CHECK-make-cvector) ((CHECK-make-cvector id ...) make-cvector))
(extend-syntax (CHECK-cvector) ((CHECK-cvector id ...) cvector))
(extend-syntax (CHECK-cvector-length) ((CHECK-cvector-length id ...) cvector-length))
(extend-syntax (CHECK-cvector-ref) ((CHECK-cvector-ref id ...) cvector-ref))
(extend-syntax (CHECK-cvector->list) ((CHECK-cvector->list id ...) cvector->list))
(extend-syntax (CHECK-list->cvector) ((CHECK-list->cvector id ...) list->cvector))


    ; structures
(let ((arg-name (lambda (x) (if (symbol? x) x (cadr x))))
      (with-mutator? (lambda (x) (not (symbol? x)))))
  (extend-syntax (CHECK-define-const-structure)
    ((CHECK-define-const-structure (name id1 ...))
     (and (symbol? 'name)
          (andmap (lambda (x)
                    (or (symbol? x)
                        (and (pair? x)
                             (equal? (car x) '!)
                             (pair? (cdr x))
                             (symbol? (cadr x))
                             (null? (cddr x)))))
            '(id1 ...)))
     (with
       ((constructor
         (symbol-append 'make- 'name))
        (CHECK-constructor
         (symbol-append 'CHECK-make- 'name))
        (predicate
         (symbol-append 'name '?))
        (CHECK-predicate
         (symbol-append 'CHECK- 'name '?))
        ((access ...)
         (let loop ((l '(id1 ...)))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l)))
             (else (cons (symbol-append 'CHECK- 'name '- (arg-name (car l)))
                         (loop (cdr l)))))))
        ((access-proc ...)
         (let loop ((l '(id1 ...)))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l)))
             (else (cons (symbol-append 'CHECK:proc- 'name '- (arg-name (car l)))
                         (loop (cdr l)))))))
        ((access-index ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l) (+ 1 n)))
             (else (cons n (loop (cdr l) (+ 1 n)))))))
        ((n-access ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             (else (cons (symbol-append 'CHECK- 'name '- n)
                         (loop (cdr l) (+ 1 n)))))))
        ((n-access-proc ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             (else (cons (symbol-append 'CHECK:proc- 'name '- n)
                         (loop (cdr l) (+ 1 n)))))))
        ((n-access-index ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             (else (cons n (loop (cdr l) (+ 1 n)))))))
        ((assign ...)
         (let loop ((l '(id1 ...)))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l)))
             ((not (with-mutator? (car l))) (loop (cdr l)))
             (else
              (cons (symbol-append 'CHECK-set- 'name '- (arg-name (car l)) '!)
                    (loop (cdr l)))))))
        ((assign-proc ...)
         (let loop ((l '(id1 ...)))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l)))
             ((not (with-mutator? (car l))) (loop (cdr l)))
             (else
              (cons (symbol-append 'CHECK:proc-set- 'name '- (arg-name (car l)) '!)
                    (loop (cdr l)))))))
        ((assign-index ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             ((eq? '_ (arg-name (car l))) (loop (cdr l) (+ 1 n)))
             ((not (with-mutator? (car l))) (loop (cdr l) (+ 1 n)))
             (else (cons n (loop (cdr l) (+ 1 n)))))))
        ((n-assign ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             ((not (with-mutator? (car l))) (loop (cdr l) (+ 1 n)))
             (else (cons (symbol-append 'CHECK-set- 'name '- n '!)
                         (loop (cdr l) (+ 1 n)))))))
        ((n-assign-proc ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             ((not (with-mutator? (car l))) (loop (cdr l) (+ 1 n)))
             (else (cons (symbol-append 'CHECK:proc-set- 'name '- n '!)
                         (loop (cdr l) (+ 1 n)))))))
        ((n-assign-index ...)
         (let loop ((l '(id1 ...))(n 1))
           (cond
             ((null? l) '())
             ((not (with-mutator? (car l))) (loop (cdr l) (+ 1 n)))
             (else (cons n (loop (cdr l) (+ 1 n)))))))
        (nargs (length '(id1 ...)))
        (ddd '...))
       (begin
         (define-const-structure (name id1 ...) ())
         (define CHECK-constructor (lambda (id) constructor))
         (define CHECK-predicate (lambda (id) predicate))
         (extend-syntax (access) ((access id ddd) access-proc))
         ...
         (define access-proc
           (lambda (obj)
             (#2%vector-ref obj access-index)))
         ...
         (extend-syntax (assign) ((assign id ddd) assign-proc))
         ...
         (define assign-proc
           (lambda (obj newval)
             (#2%vector-set! obj assign-index newval)))
         ...
         (extend-syntax (n-access) ((n-access id ddd) n-access-proc))
         ...
         (define n-access-proc
           (lambda (obj)
             (#2%vector-ref obj n-access-index)))
         ...
         (extend-syntax (n-assign) ((n-assign id ddd) n-assign-proc))
         ...
         (define n-assign-proc
           (lambda (obj newval)
             (#2%vector-set! obj n-assign-index newval)))
         ...
         )))))



    ; Chez Scheme extensions from custom-chez.scm
(extend-syntax (CHECK-call/cc) ((CHECK-call/cc id ...) #2%call/cc))
(extend-syntax (CHECK-add1) ((CHECK-add1 id ...) #2%1+))
(extend-syntax (CHECK-sub1) ((CHECK-sub1 id ...) #2%1-))
(extend-syntax (CHECK-1+) ((CHECK-1+ id ...) #2%1+))
(extend-syntax (CHECK-1-) ((CHECK-1- id ...) #2%1-))
(extend-syntax (CHECK-ormap) ((CHECK-ormap id ...) #2%ormap))
(extend-syntax (CHECK-andmap) ((CHECK-andmap id ...) #2%andmap))
(extend-syntax (CHECK-make-list) ((CHECK-make-list id ...) #2%make-list))
(extend-syntax (CHECK-format) ((CHECK-format id ...) #2%format))
(extend-syntax (CHECK-error) ((CHECK-error id ...) #2%error))
(extend-syntax (CHECK-printf) ((CHECK-printf id ...) #2%printf))
(extend-syntax (CHECK-pretty-print) ((CHECK-pretty-print id ...) #2%pretty-print))
(extend-syntax (CHECK-gensym) ((CHECK-gensym id ...) #2%gensym))
(extend-syntax (CHECK-sort) ((CHECK-sort id ...) #2%sort))
(extend-syntax (CHECK-dynamic-wind) ((CHECK-dynamic-wind id ...) #2%dynamic-wind))
(extend-syntax (CHECK-expand-once) ((CHECK-expand-once id ...) #2%eps-expand-once))
(extend-syntax (CHECK-append!) ((CHECK-append! id ...) #2%append!))
(extend-syntax (CHECK-abort) ((CHECK-abort id ...) abort))
